package dev.fyloz.trial.colorrecipesexplorer.service

import dev.fyloz.trial.colorrecipesexplorer.exception.CannotDeleteEntityException
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityAlreadyExistsException
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityNotFoundException
import dev.fyloz.trial.colorrecipesexplorer.model.*
import dev.fyloz.trial.colorrecipesexplorer.repository.MixTypeRepository
import org.springframework.context.annotation.Lazy
import org.springframework.stereotype.Service

interface MixTypeService : NamedModelService<MixType, MixTypeRepository> {
    /** Checks if a [MixType] with the given [name] and [materialType] exists. */
    fun existsByNameAndMaterialType(name: String, materialType: MaterialType): Boolean

    /** Gets the mix type with the given [material]. */
    fun getByMaterial(material: Material): MixType

    /** Gets the [MixType] with the given [name] and [materialType]. */
    fun getByNameAndMaterialType(name: String, materialType: MaterialType): MixType

    /** Returns a [MixType] for the given [name] and [materialType]. If a mix type with these does not already exists, it will be created. */
    fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialType): MixType

    /** Returns a new and persisted [MixType] with the given [name] and [materialType]. */
    fun saveForNameAndMaterialType(name: String, materialType: MaterialType): MixType

    /** Returns the given [mixType] updated with the given [name] and [materialType]. */
    fun updateForNameAndMaterialType(mixType: MixType, name: String, materialType: MaterialType): MixType
}

@Service
class MixTypeServiceImpl(
    mixTypeRepository: MixTypeRepository,
    @Lazy val materialService: MaterialService,
    @Lazy val mixService: MixService
) :
    AbstractNamedModelService<MixType, MixTypeRepository>(mixTypeRepository), MixTypeService {
    override fun existsByNameAndMaterialType(name: String, materialType: MaterialType): Boolean =
        repository.existsByNameAndMaterialType(name, materialType)

    override fun getByMaterial(material: Material): MixType =
        repository.findByMaterial(material) ?: throw EntityNotFoundException(material.name)

    override fun getByNameAndMaterialType(name: String, materialType: MaterialType): MixType =
        repository.findByNameAndMaterialType(name, materialType)
            ?: throw EntityNotFoundException("$name/${materialType.name}")

    override fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialType): MixType =
        if (existsByNameAndMaterialType(name, materialType))
            getByNameAndMaterialType(name, materialType)
        else
            saveForNameAndMaterialType(name, materialType)

    override fun save(entity: MixType): MixType {
        if (materialService.existsByName(entity.name))
            throw EntityAlreadyExistsException(entity.name)
        return super.save(entity)
    }

    override fun saveForNameAndMaterialType(name: String, materialType: MaterialType): MixType =
        save(
            mixType(
                name = name,
                material = material(
                    name = name,
                    inventoryQuantity = Float.MIN_VALUE,
                    isMixType = true,
                    materialType = materialType
                )
            )
        )

    override fun updateForNameAndMaterialType(mixType: MixType, name: String, materialType: MaterialType): MixType =
        update(mixType.apply {
            this.name = name
            material.name = name
            material.materialType = materialType
        })

    override fun deleteById(id: Long) {
        if (!repository.canBeDeleted(id)) throw CannotDeleteEntityException(id)
        super.deleteById(id)
    }
}
