package dev.fyloz.trial.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityAlreadyExistsException
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityNotFoundException
import dev.fyloz.trial.colorrecipesexplorer.model.EntityDto
import dev.fyloz.trial.colorrecipesexplorer.model.Model
import dev.fyloz.trial.colorrecipesexplorer.model.NamedModel
import dev.fyloz.trial.colorrecipesexplorer.repository.NamedJpaRepository
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.data.jpa.repository.JpaRepository
import java.util.*
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue
import dev.fyloz.trial.colorrecipesexplorer.service.AbstractServiceTest as AbstractServiceTest1

abstract class AbstractServiceTest<E, S : Service<E, *>, R : JpaRepository<E, *>> {
    protected abstract val repository: R
    protected abstract val service: S

    protected abstract val entity: E
    protected abstract val anotherEntity: E

    protected val entityList: List<E>
        get() = listOf(
            entity,
            anotherEntity
        )

    @AfterEach
    open fun afterEach() {
        reset(repository, service)
    }

    // getAll()

    @Test
    open fun `getAll() returns all available entities`() {
        whenever(repository.findAll()).doReturn(entityList)

        val found = service.getAll()

        assertEquals(entityList, found)
    }

    @Test
    open fun `getAll() returns empty list when there is no entities`() {
        whenever(repository.findAll()).doReturn(listOf())

        val found = service.getAll()

        assertTrue { found.isEmpty() }
    }

    // save()

    @Test
    open fun `save() saves in the repository and returns the saved value`() {
        whenever(repository.save(entity)).doReturn(entity)

        val found = service.save(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    // update()

    @Test
    open fun `update() saves in the repository and returns the updated value`() {
        whenever(repository.save(entity)).doReturn(entity)

        val found = service.update(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    // delete()

    @Test
    open fun `delete() deletes in the repository`() {
        service.delete(entity)

        verify(repository).delete(entity)
    }
}

abstract class AbstractModelServiceTest<E : Model, S : ModelService<E, *>, R : JpaRepository<E, Long>> :
    AbstractServiceTest1<E, S, R>() {

    // existsById()

    @Test
    open fun `existsById() returns true when an entity with the given id exists in the repository`() {
        whenever(repository.existsById(entity.id!!)).doReturn(true)

        val found = service.existsById(entity.id!!)

        assertTrue(found)
    }

    @Test
    open fun `existsById() returns false when no entity with the given id exists in the repository`() {
        whenever(repository.existsById(entity.id!!)).doReturn(false)

        val found = service.existsById(entity.id!!)

        assertFalse(found)
    }

    // getById()

    @Test
    open fun `getById() returns the entity with the given id from the repository`() {
        whenever(repository.findById(entity.id!!)).doReturn(Optional.of(entity))

        val found = service.getById(entity.id!!)

        assertEquals(entity, found)
    }

    @Test
    open fun `getById() throws EntityNotFoundException when no entity with the given id exists in the repository`() {
        whenever(repository.findById(entity.id!!)).doReturn(Optional.empty())

        val exception = assertThrows<EntityNotFoundException> { service.getById(entity.id!!) }
        assertTrue(exception.value is Long)
        assertEquals(entity.id, exception.value as Long)
    }

    // save()

    @Test
    open fun `save() throws EntityAlreadyExistsException when an entity with the given id exists in the repository`() {
        doReturn(true).whenever(repository).existsById(entity.id!!)

        val exception = assertThrows<EntityAlreadyExistsException> { service.save(entity) }
        assertTrue(exception.value is Long)
        assertEquals(entity.id, exception.value as Long)
    }

    // update()

    @Test
    override fun `update() saves in the repository and returns the updated value`() {
        whenever(repository.save(entity)).doReturn(entity)
        doReturn(true).whenever(service).existsById(entity.id!!)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.update(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    @Test
    open fun `update() throws EntityNotFoundException when no entity with the given id exists in the repository`() {
        doReturn(false).whenever(service).existsById(entity.id!!)

        val exception = assertThrows<EntityNotFoundException> { service.update(entity) }
        assertTrue(exception.value is Long)
        assertEquals(entity.id, exception.value as Long)
    }

    // deleteById()

    @Test
    open fun `deleteById() deletes the entity with the given id in the repository`() {
        doReturn(entity).whenever(service).getById(entity.id!!)

        service.deleteById(entity.id!!)

        verify(repository).delete(entity)
    }
}

abstract class AbstractNamedModelServiceTest<E : NamedModel, S : NamedModelService<E, *>, R : NamedJpaRepository<E>> :
    AbstractModelServiceTest<E, S, R>() {
    protected abstract val entityWithEntityName: E

    // existsByName()

    @Test
    open fun `existsByName() returns true when an entity with the given name exists`() {
        whenever(repository.existsByName(entity.name)).doReturn(true)

        val found = service.existsByName(entity.name)

        assertTrue(found)
    }

    @Test
    open fun `existsByName() returns false when no entity with the given name exists`() {
        whenever(repository.existsByName(entity.name)).doReturn(false)

        val found = service.existsByName(entity.name)

        assertFalse(found)
    }

    // getByName()

    @Test
    open fun `getByName() returns the entity with the given name`() {
        whenever(repository.findByName(entity.name)).doReturn(entity)

        val found = service.getByName(entity.name)

        assertEquals(entity, found)
    }

    @Test
    open fun `getByName() throws EntityNotFoundException when no entity with the given name exists`() {
        whenever(repository.findByName(entity.name)).doReturn(null)

        val exception = assertThrows<EntityNotFoundException> { service.getByName(entity.name) }
        assertEquals(entity.name, exception.value)
    }

    // save()

    @Test
    open fun `save() throws EntityAlreadyExistsException when an entity with the given name exists`() {
        doReturn(true).whenever(service).existsByName(entity.name)

        val exception = assertThrows<EntityAlreadyExistsException> { service.save(entity) }
        assertEquals(entity.name, exception.value)
    }

    // update()

    @Test
    override fun `update() saves in the repository and returns the updated value`() {
        whenever(repository.save(entity)).doReturn(entity)
        whenever(repository.findByName(entity.name)).doReturn(null)
        doReturn(true).whenever(service).existsById(entity.id!!)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val found = service.update(entity)

        verify(repository).save(entity)
        assertEquals(entity, found)
    }

    @Test
    override fun `update() throws EntityNotFoundException when no entity with the given id exists in the repository`() {
        whenever(repository.findByName(entity.name)).doReturn(null)
        doReturn(false).whenever(service).existsById(entity.id!!)

        val exception = assertThrows<EntityNotFoundException> { service.update(entity) }
        assertTrue(exception.value is Long)
        assertEquals(entity.id, exception.value as Long)

    }

    @Test
    open fun `update() throws EntityAlreadyExistsException when an entity with the updated name exists`() {
        whenever(repository.findByName(entity.name)).doReturn(entityWithEntityName)
        doReturn(entity).whenever(service).getById(entity.id!!)

        val exception = assertThrows<EntityAlreadyExistsException> { service.update(entity) }
        assertEquals(entity.name, exception.value)
    }

    // deleteByName()

    @Test
    open fun `deleteByName() deletes the entity with the given name in the repository`() {
        service.deleteByName(entity.name)

        verify(repository).deleteByName(entity.name)
    }
}

interface ExternalModelServiceTest {
    fun `save(dto) calls and returns save() with the created entity`()
    fun `update(dto) calls and returns update() with the created entity`()
}

// ==== IMPLEMENTATIONS FOR EXTERNAL SERVICES ====
// Lots of code duplication but I don't have a better solution for now
abstract class AbstractExternalModelServiceTest<E : Model, N : EntityDto<E>, U : EntityDto<E>, S : ExternalModelService<E, N, U, *>, R : JpaRepository<E, Long>> :
    AbstractModelServiceTest<E, S, R>(), ExternalModelServiceTest {
    protected abstract val entitySaveDto: N
    protected abstract val entityUpdateDto: U

    @AfterEach
    override fun afterEach() {
        reset(entitySaveDto, entityUpdateDto)
        super.afterEach()
    }
}

abstract class AbstractExternalNamedModelServiceTest<E : NamedModel, N : EntityDto<E>, U : EntityDto<E>, S : ExternalNamedModelService<E, N, U, *>, R : NamedJpaRepository<E>> :
    AbstractNamedModelServiceTest<E, S, R>(), ExternalModelServiceTest {
    protected abstract val entitySaveDto: N
    protected abstract val entityUpdateDto: U

    @AfterEach
    override fun afterEach() {
        reset(entitySaveDto, entityUpdateDto)
        super.afterEach()
    }
}

fun <E : Model, N : EntityDto<E>> withBaseSaveDtoTest(
    entity: E,
    entitySaveDto: N,
    service: ExternalService<E, N, *, *>,
    saveMockMatcher: () -> E = { entity },
    op: () -> Unit = {}
) {
    doReturn(entity).whenever(service).save(saveMockMatcher())
    doReturn(entity).whenever(entitySaveDto).toEntity()

    val found = service.save(entitySaveDto)

    verify(service).save(saveMockMatcher())
    assertEquals(entity, found)

    op()
}

fun <E : Model, U : EntityDto<E>> withBaseUpdateDtoTest(
    entity: E,
    entityUpdateDto: U,
    service: ExternalModelService<E, *, U, *>,
    updateMockMatcher: () -> E,
    op: E.() -> Unit = {}
) {
    doAnswer { it.arguments[0] }.whenever(service).update(updateMockMatcher())
    doReturn(entity).whenever(entityUpdateDto).toEntity()
    doReturn(entity).whenever(service).getById(entity.id!!)
    doReturn(true).whenever(service).existsById(entity.id!!)

    val found = service.update(entityUpdateDto)

    verify(service).update(updateMockMatcher())
    assertEquals(entity, found)

    found.op()
}
