package dev.fyloz.trial.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.trial.colorrecipesexplorer.exception.LowQuantitiesException
import dev.fyloz.trial.colorrecipesexplorer.exception.LowQuantityException
import dev.fyloz.trial.colorrecipesexplorer.model.MaterialQuantityDto
import dev.fyloz.trial.colorrecipesexplorer.model.material
import dev.fyloz.trial.colorrecipesexplorer.model.materialQuantityDto
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class InventoryServiceTest {
    private val materialService: MaterialService = mock()
    private val service = spy(InventoryServiceImpl(materialService))

    @AfterEach
    fun afterEach() {
        reset(materialService, service)
    }

    // add()

    @Test
    fun `add(materialQuantities) calls add() for each MaterialQuantityDto`() {
        val materialQuantities = listOf(
            materialQuantityDto(materialId = 1, quantity = 1234f),
            materialQuantityDto(materialId = 2, quantity = 2345f),
            materialQuantityDto(materialId = 3, quantity = 3456f),
            materialQuantityDto(materialId = 4, quantity = 4567f)
        )

        service.add(materialQuantities)

        materialQuantities.forEach {
            verify(service).add(it)
        }
    }

    @Test
    fun `add(materialQuantity) updates material's quantity`() {
        withGivenQuantities(0f, 1000f) {
            service.add(it)

            verify(materialService).updateQuantity(
                argThat { this.id == it.material },
                eq(it.quantity)
            )
        }
    }

    // deduct()

    @Test
    fun `deduct(materialQuantities) calls deduct() for each MaterialQuantityDto`() {
        val materialQuantities = listOf(
            materialQuantityDto(materialId = 1, quantity = 1234f),
            materialQuantityDto(materialId = 2, quantity = 2345f),
            materialQuantityDto(materialId = 3, quantity = 3456f),
            materialQuantityDto(materialId = 4, quantity = 4567f)
        )

        service.deduct(materialQuantities)

        materialQuantities.forEach {
            verify(service).deduct(it)
        }
    }

    @Test
    fun `deduct(materialQuantities) throws LowQuantitiesException when there is not enough inventory of a given material`() {
        val materialQuantities = listOf(
            materialQuantityDto(materialId = 1, quantity = 1234f),
            materialQuantityDto(materialId = 2, quantity = 2345f),
            materialQuantityDto(materialId = 3, quantity = 3456f),
            materialQuantityDto(materialId = 4, quantity = 4567f)
        )
        val inventoryQuantity = 3000f
        val lowQuantities = materialQuantities.filter { it.quantity > inventoryQuantity }

        materialQuantities.forEach {
            withGivenQuantities(inventoryQuantity, it)
        }

        val exception = assertThrows<LowQuantitiesException> { service.deduct(materialQuantities) }
        assertTrue { exception.materialQuantities.containsAll(lowQuantities) }
    }

    @Test
    fun `deduct(materialQuantity) updates material's quantity`() {
        withGivenQuantities(5000f, 1000f) {
            service.deduct(it)

            verify(materialService).updateQuantity(
                argThat { this.id == it.material },
                eq(-it.quantity)
            )
        }
    }

    @Test
    fun `deduct(materialQuantity) throws LowQuantityException when there is not enough inventory of the given material`() {
        withGivenQuantities(0f, 1000f) {
            val exception = assertThrows<LowQuantityException> { service.deduct(it) }
            assertEquals(it, exception.materialQuantity)
        }
    }

    private fun withGivenQuantities(
        inventory: Float,
        deductedQuantity: Float,
        materialId: Long = 0L,
        test: (MaterialQuantityDto) -> Unit = {}
    ) {
        val materialQuantity = materialQuantityDto(materialId = materialId, quantity = deductedQuantity)

        withGivenQuantities(inventory, materialQuantity, test)
    }

    private fun withGivenQuantities(
        inventory: Float,
        materialQuantity: MaterialQuantityDto,
        test: (MaterialQuantityDto) -> Unit = {}
    ) {
        val material = material(id = materialQuantity.material, inventoryQuantity = inventory)

        whenever(materialService.getById(material.id!!)).doReturn(material)

        test(materialQuantity)
    }
}
